/* unix.cpp
 * ports of parts of STk's unix.c
 */

/* Explanation of some Windows programming:
 *
 * TCHAR is either char or wchar_t, depending on
 * if you are compiling for ascii or unicode.
 *
 * _T(x) and TEXT(x) is either x or Lx, to make a string
 * a const char* or const wchar_t*. It works for single characters too,
 * is not needed for chars < 127.
 *
 * Stuff like _tcslen becomes either strlen or wcslen.
 * Microsoft also adds a bunch of reasonable _str routines,
 * like _strnicmp for a length limited case insensitive string compare.
 * _tcsnicmp and _wcsnicmp too.
 *
 * Unless STk, Tcl, and Tk are TCHAR-ized, this is pretty useless and
 * will only work in ascii mode.
 *
 * Windows NT defines functions like GetUserNameA and GetUserNameW and
 * depending on the UNICODE preprocess symbol, defines GetUserName to
 * one of them. Windows 95 implements very few "W" functions.
 *
 * Canonical paths on Windows notoriously use backslashes, but all functions
 * accept forward slashes just as well, but most command line utils and programs
 * do not. (For example, to use forward slashes with dir requires quotes, and
 * even then only seems to work for directories, not files.)
 *
 * Windows 95 full paths are limited to 260 characters, which is defined to be MAX_PATH.
 * Windows NT has the same limit "by default", but if you pass it full paths prefixed
 * by \\?\ or \\?\UNC\, the limit can be surpassed.
 * eg: \\?\c:\winnt\system32
 * \\?\UNC\jayk3\c\winnt
 * This is extremely ugly, but at least it is 1) compatible, and 2) allows unlimited
 * length paths.
 *
 * GetFullPathName "computes" full paths. It is never limited to MAX_PATH. It simply
 * converts all slashes to back slashes, folds slash runs longer than 3,
 * prepends the "current directory" or "current drive"
 * to non absolute paths, and processes any . or .. elements. I believe it never hits
 * the disk.
 *
 * I think NT sets up HOMEDRIVE and HOMEPATH environment variables by default,
 * but not HOME. But you are free to add HOME yourself.
 *
 * typedef TCHAR* LPTSTR;
 * typedef const TCHAR* LPCTSTR;
 *
 * This is dubious, esp. the "L", but I can live with it.
 */

#include <math.h> // had some C/C++ problems
#include <assert.h>
#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <tchar.h>
#include "stk.h"
#include "alloca.h"

static void BackwardToForwardSlashes(char* s)
{
	if (s)
	{
		while (s = strchr(s, '\\'))
		{
			*s = '/';
		}
	}
}

extern "C" void absolute(const char* s, char* pathname)
{
	GetFullPathNameA(s, MAX_PATH, pathname, NULL);
	BackwardToForwardSlashes(pathname);
}

extern "C" char* tilde_expand(char* name, char* result)
{
	assert(name && result);
	int ch0 = name[0];
	result[0] = 0;
	if (ch0 != '~')
	{
		strcat(result, name);
	}
	else
	{
		int ch1 = name[1];
		if (ch1 == '/' || ch1 == '\\' || ch1 == 0)
		{
			char   home[MAX_PATH], homedrive[MAX_PATH], homepath[MAX_PATH];
			home[0] = homedrive[0] = homepath[0] = 0;

			DWORD homelen = GetEnvironmentVariableA("HOME", home, MAX_PATH);
			DWORD homedrivelen = GetEnvironmentVariableA("HOMEDRIVE", homedrive, MAX_PATH);
			DWORD homepathlen = GetEnvironmentVariableA("HOMEPATH", homepath, MAX_PATH);

			if (homelen != 0 && homelen < MAX_PATH && home[0])
			{
				strcat(result, home);
				strcat(result, name+1);
			}
			else if (
				homedrivelen && homedrivelen < MAX_PATH && homepath[0]
				&&
				homepathlen  && homepathlen  < MAX_PATH && homedrive[0])
			{
				strcat(result, homedrive);
				strcat(result, homepath);
				strcat(result, name+1);
			}
			else
			{
				Err(TEXT("couldn't find HOME[DRIVE|PATH] environment for expanding "), STk_makestring(name));
			}
		}
		else
		{
			/* check for ~current_user... but can't do ~any_user */
			CHAR user[MAX_PATH]; /* can't seem to find the needed length */
			DWORD size = MAX_PATH;
			if (GetUserNameA(user, &size) && size < MAX_PATH && user[0])
			{
				int len = strlen(user);
				if (0 == _strnicmp(name+1, user, len)
					&& ((name+1)[len] == '/'
						|| (user+1)[len] == '\\'
						|| (user+1)[len] == 0
						)
					)
				{
					/* remove user name and call recursively. LAZY */
					LPSTR without_user = (LPTSTR)alloca((strlen(name)+1)*sizeof(*without_user));
					without_user[0] = '~';
					strcpy(without_user+1, name+1+len);
					return tilde_expand(without_user, result);
				}
				Err(TEXT("Can't lookup non current user's $HOME on Windows."), STk_makestring(name));
			}
			Err(TEXT("GetUserName() failed or user name is too long or empty."), STk_makestring(name));
		}
	}
	BackwardToForwardSlashes(result);
	return result;
}

#if 0
/* !<symlink>F:\winnt\_DEFAULT.PIF
*/
/* try to support Cygwin32 b17.1 symbolic links. */
extern "C" int readlink(LPTSTR result, LPCTSTR src, int result_size)
{
	int lnk = _open(src, _O_BINARY|_O_RDONLY|_O_SEQUENTIAL);
	if (lnk == -1)
	{
		return -1;
	}
	long len = _filelength(lnk);
	static const char magic[] = "!<symlink>";
	const int magic_length = strlen(magic);
	if (len < (magic_length+1)) // assume link data is > 0
	{
		_close(lnk);
		return -1;
	}
	if (len > (magic_length+MAX_PATH+3))
	{ /* allow trailing [\n\r\0] */
		_close(lnk);
		return -1;
	}
	char* sym_buf = (char*)alloca((len+1)*sizeof(*sym_buf));
	if (len != _read(lnk, sym_buf, len))
	{
		return -1;
	}
	_close(lnk);
	if (0 != memcmp(magic, sym_buf, magic_length))
	{
		return -1; // not a link, expensive to determine.
	}
	sym_buf[len] = 0;
	while (sym_buf[len] == 0 || sym_buf[len] == 13 || sym_buf[len] == 10)
	{
		len--;
	}
	len++;
	sym_buf[len] = 0;
	strcpy(result, sym_buf);
	return len;
}
#endif
